{
  a Z-80 assembly language compiler for ZX-SPECTRUM (48/128/512/1024)

  DESCRIPT: macros handler.

                                 was written by SCREW/OHG/CM
                                   all rights reserved
}
Unit C_Z80Macros;

INTERFACE

Const max_Macros=50;     { max of macros }
      max_Macro_Len=20;  { max of macros text lines }
      max_Macro_Name=16; { max length of macros name }
      max_Macro_Line=50; { max length of macros text lines }
Type Macros=Record
             name:string[max_Macro_Name];
             params:string;
             text:array[1..max_Macro_len] of string[max_Macro_Line];
             lines:integer;
            end;
     mac_table=Record
                macs:array[1..max_Macros] of Macros;
                total:integer;
                last_found:integer;
                last_name:string[max_Macro_Name]
               end;
      mac_Ptr=^mac_table;

Procedure Macro_Init(var M:mac_Ptr);
Procedure Macro_Reinit(var M:mac_Ptr);
Procedure Macro_Deinit(var M:mac_Ptr);
Function isMacros(M:mac_Ptr; name:string):boolean;
Procedure Macro_Add(var M:mac_Ptr; text:string);
Function Replace(s,s1,s2:string):string;
Function Macro_Get(var M:mac_Ptr; text:string):string;

Var mac:mac_Ptr;
    is_next_macro:boolean; {if next macro-line present}
      is_next_add:boolean; {if next macro-line addition}
        curr_numb:integer; { current line number (in macros) }
      curr_params:string;  { current parameters (in macros) }

IMPLEMENTATION
Uses C_ErrHandle, { using error handler }
     C_Z80Codes;  { using GetWord }
{$I C_Z80ERR.pas} {include errors definition }


{***************************************************************************
                      ALL CODE IMPLEMENTATIONS
***************************************************************************}

Procedure Macro_Init(var M:mac_Ptr);
Begin
 new(M);
 M^.total:=0;
 M^.last_found:=1;
 is_next_macro:=false;
 is_next_add:=false
end;

Procedure Macro_Reinit(var M:mac_Ptr);
Begin
 M^.last_found:=1;
 is_next_macro:=false;
 is_next_add:=false
end;

Procedure Macro_Deinit(var M:mac_Ptr);
Begin
 dispose(M)
end;
{
Procedure Macro_Reset(var M:mac_Ptr);
Begin
 M^.total:=0
end;
 }
Function isMacros(M:mac_Ptr; name:string):boolean;
Var i:integer;
Begin
 isMacros:=false;
 for i:=1 to M^.total do
  if M^.macs[i].name=name then Begin
                                isMacros:=true;
                                M^.last_found:=i;
                                break
                               end
end;

Procedure Macro_Add(var M:mac_Ptr; text:string);
var num,i,len:integer;
    name:string;
Begin
 if not is_next_add then Begin
                          i:=1; name:=GetWord(text,i);
                          GetWord(text,i); {skip 'macro'}
                          text:=copy(text,i,length(text)-i+1);
                          { macro add init }
                          if not isMacros(M,name) then M^.last_found:=M^.total+1;
                          num:=M^.last_found;
                          if num>max_Macros then Begin
                                                  Error(TOO_MANY_MACROS);
                                                  exit
                                                 end;
                          M^.macs[num].name:=name;
                          M^.macs[num].params:=text;
                          M^.macs[num].lines:=0;
                          is_next_add:=true
                         end
else Begin
      { add next line in macros }
      num:=M^.last_found;
      i:=1;
      name:=GetWord(text,i);
      if name='' then exit; {no ADD for emtpy line}
      if name = 'ENDM' then Begin
                             is_next_add:=false;
                             if num>M^.total then M^.total:=num;
                             exit
                            end;
      if M^.macs[num].lines>=max_macro_len then Begin
                                                 Error(MACROS_IS_TOO_LARGE);
                                                 exit
                                                end;
      inc(M^.macs[num].lines);
      i:=1; len:=length(text); name:='';
      while(i<=len) do
       Begin
        name:=name+text[i];
        if text[i]<>' ' then inc(i) else
         while(text[i]=' ') and (i<=len) do inc(i)
       end;
      M^.macs[num].text[ M^.macs[num].lines ]:=name
     end
end;

Function Replace(s,s1,s2:string):string;
Var i,j,len,len1:integer;
      prs:boolean;
Begin
 len:=length(s);
 len1:=length(s1);
 i:=1;
 while(i<=len-len1+1) do
  Begin
   prs:=true;
   for j:=0 to len1-1 do if s[i+j]<>s1[j+1] then Begin
                                                  prs:=false;
                                                  break
                                                 end;
   if prs then Begin
                s:=copy(s,1,i-1)+s2+copy(s,i+len1,len-i-len1+1);
                len:=length(s);
               end;
   inc(i);
   if i>255 then Begin
                  Error(INFINITE_RECURSE);
                  exit
                 end
  end;
 Replace:=s
end;

Function Macro_Get(var M:mac_Ptr; text:string):string;
var num,i,j:integer;
    name,s1,s2,s3:string;
    p1,p2,cond:string;
          skip:boolean;
Begin
 if not is_next_macro then Begin
                            i:=1; name:=GetWord(text,i);
                            { macro get init }
                            if not isMacros(M,name) then Begin
                                                  {       Error(SYNTAX_ERROR);
                                                   }      exit
                                                        end;
                             { get all parameters }
                            curr_params:='';
                            repeat
                             name:=GetWord(text,i);
                             curr_params:=curr_params+name+' ';
                            until name='';
                             { adjust variables }
                            curr_numb:=1;
                            is_next_macro:=true;
                          end
else Begin
      { get next line from macros }
      num:=M^.last_found;
      i:=1;
      if curr_numb>M^.macs[num].lines then Begin
                                            is_next_macro:=false;
                                            Macro_Get:='';
                                            exit
                                           end;
      name:=M^.macs[num].text[ curr_numb ];
      {if condition}
      i:=1;
      if GetWord(name,i)='ELSE' then Begin
                                      is_next_macro:=false;
                                      Macro_Get:='';
                                      exit
                                     end;
      i:=1;
      if GetWord(name,i)='IF' then
       Begin
        name:=Replace(name,'==',' equal ');
        name:=Replace(name,'!=',' unequal ');
        p1:=GetWord(name,i);
        cond:=GetWord(name,i);
        p2:=GetWord(name,i);
        i:=1; j:=1;
        {search params values}
        repeat
        s1:=GetWord(curr_params,i);
        s2:=GetWord(M^.macs[num].params,j);
        if ( (s1='') and (s2<>'') ) then Begin
                                          Error(MACROS_PARAMETER_MISS);
                                          exit;
                                         end;
        if ( (s2='') and (s1<>'') ) then Begin
                                          Error(TOO_MANY_MACROS_PARAMETERS);
                                          exit
                                         end;
{        if s2=p1 then p1:=s1;
        if s2=p2 then p2:=s1  }
        p1:=replace(p1,s2,s1);
        p2:=replace(p2,s2,s1)
        until s2='';
        {high/low part of register using}
        High_Low(p1);
        High_Low(p2);
        skip:=false;
        if (cond='equal') and (p1<>p2) then skip:=true
       else
        if (cond='unequal') and (p1=p2) then skip:=true
       else
        if (cond<>'equal') and (cond<>'unequal') then Begin
                                                       Error(SYNTAX_ERROR);
                                                       exit
                                                      end;
        if skip then {go to else}
                      repeat
                       i:=1;
                       inc(curr_numb);
                       s3:=getWord(M^.macs[num].text[ curr_numb ] ,i)
                      until (s3='ELSE') or (curr_numb>M^.macs[num].lines);
        inc(curr_numb);
        Macro_Get:=Macro_Get(M,text);
        exit
       end; { IF end}


      {macro use}
      i:=1; j:=1;
      repeat
       s1:=GetWord(curr_params,i);
       s2:=GetWord(M^.macs[num].params,j);
       if ( (s1='') and (s2<>'') ) then Begin
                                         Error(MACROS_PARAMETER_MISS);
                                         exit;
                                        end;
       if ( (s2='') and (s1<>'') ) then Begin
                                         Error(TOO_MANY_MACROS_PARAMETERS);
                                         exit
                                        end;
       { in name change all s2 to s1 }
       if s2<>'' then
                      name:=Replace(name,s2,s1)
      until s2='';
      inc(curr_numb);
      Macro_Get:=name
     end
end;


end.